import numpy as np
import slicer.utils as utils
from gcode2contour.gcode2contour import Contour, Position
from scipy.interpolate import interpn


class Precision:
    p_f_p = 3  # floating point accuracy for paths
    e_f_p = 5  # floating point accuracy for extrusion
prec = Precision()


class ContourGenerator:

    def __init__(self, settings, surfs, contour_normals=False):
        """
        make contours from polyline objects
        settings: slicer.ParsedSettings object
        surfs: list of surfaces used to slice the model

        Note: surfs are needed to calculate layer heights and normals
          for non-planar contours. Surfs must not be 'pv.PolyData.clean'ed
          since surf points must be rectilinear for normal interpolation
        """
        self.base_layer_h = settings.layer_height
        self.config_extrusion_modifier = settings.extrusion_modifier  # default value
        self.extrusion_modifier = settings.extrusion_modifier  # changed for brim
        self.filament_diam = settings.filament_diam
        self.extrusion_width = settings.extrusion_width
        self.contour_normals = contour_normals

        # For plane normals to work, all surf points must be described a set of rectilinear points
        self.surfs = surfs
        self.layer_proximity_obj = [None] * len(surfs)
        self.surf_points = []
        self.surf_normals = []
        for surf in surfs:
            xs = np.unique(surf.points[:,0])
            ys = np.unique(surf.points[:,1])
            self.surf_points.append((xs, ys))
            self.surf_normals.append(surf.point_normals.reshape((len(ys), len(xs), 3)).swapaxes(0,1))

    def calc_extrusion(self, segment, indices):
        """
        Calculates the required extrusion amount between two points, modifies the cumulative total
        """
        segment_dist = utils.calc_line_dist(segment[0], segment[1])
        segment_area = self.calc_segment_area(indices)
        return 4 * segment_dist * segment_area/ (np.pi * (self.filament_diam ** 2)) * self.extrusion_modifier

    def calc_segment_area(self, indices):
        """
        Calculates the area of an extrusion path (depends on relative layer height)
        """
        if isinstance(self.curr_path_layer_h, float):
            rel_h = self.curr_path_layer_h
        else:
            rel_h = self.calc_layer_height_segment(indices)
        return ((self.extrusion_width - rel_h) * rel_h) + np.pi * (rel_h / 2)**2

    def calc_layer_height_segment(self, indices):
        """
        Returns the mean layer height for a polyline segment
        """
        return np.mean([self.curr_path_layer_h[indices[0]], self.curr_path_layer_h[indices[1]]])

    def calc_layer_height_all(self, points, layer):
        """
        Queries the closest point on the surface below for each point on the path
        """
        if layer == 0:
            self.curr_path_layer_h = self.base_layer_h
        else:
            self.curr_path_layer_h = utils.closest_dist(self.layer_proximity_obj[layer], points)

    def points2contour(self, points, layer_i=None):
        """
        Turns an array of xyz points into contours
        """
        c = Contour([Position(*pnt) for pnt in points],0)
        for p, pnt in enumerate(points):
            if p == 0:
                continue
            ext = self.calc_extrusion([points[p-1], pnt], [p-1, p])
            c.Ext.append(c.Ext[-1] + ext)
            if layer_i is not None and self.contour_normals:
                c.normals = self.get_contour_normals(c, layer_i)
        return c

    def paths_to_contours(self, paths, layer):
        """
        Generates contours for a list of paths
        Includes travel movements
        """
        contours = []
        if np.size(paths) == 0:
            return contours

        # convert to contours
        for path in paths:
            self.calc_layer_height_all(path, layer)
            contours.append(self.points2contour(path, layer))
        return contours

    def gen_ext_shell_contours(self, pl):
        """
        Generates contours for external shells for PolyLines object pl
        """
        return self.paths_to_contours(pl.polylines_3d[pl.ext_shells], pl.layer_i)

    def gen_int_shell_contours(self, pl):
        """
        Generates contours for internal shells, followed by intermediate shells
          for PolyLines object pl
        """
        int_shells = pl.polylines_3d[pl.int_shells]
        inter_shells = pl.polylines_3d[pl.inter_shells]
        c1 = self.paths_to_contours(int_shells, pl.layer_i)
        c2 = self.paths_to_contours(inter_shells, pl.layer_i)
        return c1 + c2

    def gen_infill_contours(self, pl):
        """
        Generates contours for shells, including special cases for outer and inner shells
        """
        return self.paths_to_contours(pl.polylines_3d[pl.infill], pl.layer_i)

    def gen_adhesion_contours(self, pl):
        """
        Generates contours for bed adhesion (eg brim) if it exists for Polyline object pl
        """
        self.extrusion_modifier = 1.1
        cs = self.paths_to_contours(pl.polylines_3d[pl.adhesion], pl.layer_i)
        self.extrusion_modifier = self.config_extrusion_modifier
        return cs

    def get_contour_normals(self, c, layer):
        """
        Returns the normals of contour c interpolated from surf normals
        """
        return interpn(self.surf_points[layer], self.surf_normals[layer], np.array(c.pos)[:,:2])

    def polylines2contours(self, pl):
        """
        Generate contours for a PolyLines object pl
        """
        if self.layer_proximity_obj[pl.layer_i] is None:
            self.layer_proximity_obj[pl.layer_i] = utils.create_prox_obj(self.surfs[pl.layer_i-1])
        cs = self.gen_int_shell_contours(pl)
        cs += self.gen_ext_shell_contours(pl)
        cs += self.gen_infill_contours(pl)
        return cs