import numpy as np
import slicer.utils as utils
from .Contour import prec


class Infill:

    def __init__(self, density, bounding_box, fill_thresh, min_spacing, resample_res):
        self.x_min, self.y_min = bounding_box[0, 0], bounding_box[0, 1]
        self.x_max, self.y_max = bounding_box[1, 0], bounding_box[1, 1]
        self.min_spacing = min_spacing
        self.density = density
        self.fill_thresh = fill_thresh
        self.resample_res = resample_res
        self.width, self.height = self.bb_dims()
        self.infill_arr = list()
        self.filling = self.can_fill_bb()

    def update_infill_arr(self, points):
        """
        Add to the infill array
        """
        if self.resample_res is not None:
            points = utils.resample_path(points, self.resample_res).tolist()  # resample for z integrity
        self.infill_arr.extend(np.round(points, prec.p_f_p))

    def get_infill_arr(self):
        """
        Return the infill paths as a numpy array
        """
        infill_arr_np = np.array([p for p in self.infill_arr], dtype=object, ndmin=3)
        return infill_arr_np if infill_arr_np.size > 0 else None

    def align_to_grid(self, spacing, base):
        """
        Extend bounding box to nearest multiple of hex pattern
        """
        def _align_to_grid(coord, _spacing):
            if coord < 0:
                aligned = ((coord - _spacing + 1) / _spacing) * _spacing
            else:
                aligned = (coord / _spacing) * _spacing
            return aligned
        self.x_min = base[0] + _align_to_grid(self.x_min - base[0], spacing[0])
        self.y_min = base[1] + _align_to_grid(self.y_min - base[1], spacing[1])

    def bb_dims(self):
        """
        Returns the dimensions of the bounding box
        """
        width = abs(self.x_max-self.x_min)
        height = abs(self.y_max-self.y_min)
        return width, height

    def can_fill_bb(self):
        """
        Checks the dimensions of a pyclipper bounding box given with keys left, right, top, bottom
        Returns true if width or height are greater than infill area threshold
        """
        width = abs(self.x_max-self.x_min)
        height = abs(self.y_max-self.y_min)
        return True if width > self.fill_thresh or height > self.fill_thresh else False


class Rectilinear(Infill):

    def __init__(self, density, angle, bounding_box, fill_thresh, min_spacing, resample_res):
        super().__init__(density, bounding_box, fill_thresh, min_spacing, resample_res)
        self.angle = angle
        self.centre = [(self.x_max + self.x_min) / 2, (self.y_max + self.y_min) / 2]
        self.min_spacing = self.min_spacing / self.density

    def fill_surface(self):
        """
        Fill a surface with rectilinear infill
        Reverses every second line for continuity
        """
        if not self.filling:
            return None
        short = self.bb_rotated_dimension_deficit()
        if self.width >= self.height:
            x_arr = self.get_dim_arr(self.x_min, self.x_max, short)
            for ix, x in enumerate(x_arr):
                line = [[x, self.y_min], [x, self.y_max]] if ix % 2 else [[x, self.y_max], [x, self.y_min]]
                reverse = False if ix % 2 else True
                line = self.extend_line_if_needed(line, reverse=reverse)
                line = utils.rotate_2d_line(line, self.centre, self.angle)
                self.update_infill_arr(line)
        else:
            y_arr = self.get_dim_arr(self.y_min, self.y_max, short)
            for iy, y in enumerate(y_arr):
                line = [[self.x_min, y], [self.x_max, y]] if iy % 2 else [[self.x_max, y], [self.x_min, y]]
                reverse = False if iy % 2 else True
                line = self.extend_line_if_needed(line, reverse=reverse)
                line = utils.rotate_2d_line(line, self.centre, self.angle)
                self.update_infill_arr(line)
        return self.get_infill_arr()

    def get_dim_arr(self, min_bound, max_bound, short):
        """
        Generate array over which lines are generated
        """
        min_bound -= short
        max_bound += short
        steps = int((max_bound - min_bound) / self.min_spacing)
        return np.linspace(min_bound, max_bound, steps)

    def bb_rotated_dimension_deficit(self):
        """
        Gets the maximum side length between the original bounding box and a rotated bounding box
        """
        p0 = [self.x_min, self.y_min]
        pa_x, pa_y = utils.rotate_2d_point(p0, self.centre, self.angle)
        short = 1.5*max(abs(pa_x - p0[0]), abs(pa_y - p0[1]))
        return short

    def extend_line_if_needed(self, line, reverse=False):
        """
        Calculates the diagonal of the bounding box; if a given line is shorter than this,
        extends the line in both directions until it is the same length as the diagonal
        """
        temp_line = line if not reverse else line[::-1]
        min_diag = np.sqrt(self.width**2 + self.height**2)
        line_dist = utils.calc_line_dist(np.array(temp_line[0]), np.array(temp_line[1]))
        diff = min_diag - line_dist
        if diff <= 0:
            return line
        dy = temp_line[1][1]-temp_line[0][1]
        dx = temp_line[1][0]-temp_line[0][0]
        m = dy/dx if dx > 0 else np.inf
        if m is not np.inf:
            temp_line[0][0] -= (diff/2)*np.cos(m)
            temp_line[0][1] -= (diff/2)*np.sin(m)
            temp_line[1][0] += (diff/2)*np.cos(m)
            temp_line[1][1] += (diff/2)*np.sin(m)
        else:  # line is vertical
            temp_line[0][1] -= diff/2
            temp_line[1][1] += diff/2
        return temp_line if not reverse else temp_line[::-1]


class Honeycomb(Infill):

    def __init__(self, density, bounding_box, fill_thresh, min_spacing, resample_res):
        super().__init__(density, bounding_box, fill_thresh, min_spacing, resample_res)
        self.distance = self.min_spacing / density
        self.hex_side = self.distance / (np.sqrt(3) / 2)
        hex_width = self.distance * 2
        hex_height = self.hex_side * 2
        pattern_height = hex_height + self.hex_side
        # hex_center = [self.hex_width / 2, self.hex_side]
        self.y_short = self.distance * np.sqrt(3) / 3
        self.x_off = min_spacing / 2
        self.y_off = self.x_off * np.sqrt(3) / 3
        centre = [(self.x_max + self.x_min) / 2, (self.y_max + self.y_min) / 2]
        self.align_to_grid([hex_width, pattern_height], centre)
        self.x = np.arange(self.x_min, self.x_max, self.distance * 2)
        self.y = np.arange(self.y_min, self.y_max, self.y_short + self.hex_side + self.y_short + self.hex_side)

    def fill_surface(self):
        """
        Fill a surface with honeycomb infill
        """
        if not self.filling:
            return None
        points = list()
        for ix, x in enumerate(self.x):
            ax = [x + self.x_off, x + self.distance - self.x_off]
            for i in range(2):
                points.reverse()
                for iy, y in enumerate(self.y):
                    points.append([ax[1], y + self.y_off])
                    points.append([ax[0], y + self.y_short - self.y_off])
                    points.append([ax[0], y + self.y_short + self.hex_side + self.y_off])
                    points.append([ax[1], y + self.y_short + self.hex_side + self.y_short - self.y_off])
                    points.append([ax[1], y + self.y_short + self.hex_side + self.y_short + self.hex_side + self.y_off])
                ax[0] = ax[0] + self.distance
                ax[1] = ax[1] + self.distance
                ax[0], ax[1] = (ax[1], ax[0])
            self.update_infill_arr(points)
            points.clear()
        return self.get_infill_arr()
