import numpy as np
import pyclipper as pc
import slicer.utils as utils
from scipy.spatial.qhull import Delaunay
from .Contour import prec


class PolyLines:
    """
    2D geometry data for a surface
    """

    def __init__(self, slice, layer_i, layer_surf):
        self.slice = slice
        self.layer_surf = layer_surf
        self.layer_i = layer_i
        self.xyz_data = self.layer_surf.points

        self.polylines = np.empty(0, dtype=object)
        self.ext_shells = np.empty(0, dtype=int)
        self.int_shells = np.empty(0, dtype=int)
        self.infill_shells = np.empty(0, dtype=int)
        self.inter_shells = np.empty(0, dtype=int)
        self.infill = np.empty(0, dtype=int)
        self.holes = np.empty(0, dtype=int)
        self.adhesion = np.empty(0, dtype=int)
        self.region_groups = np.empty(0, dtype=object)
        self.polylines_3d = np.empty(0, dtype=object)
        self.layer_de = None

    def make_collection_clipper(self, indices):
        """
        Group given paths/polylines as a clipper project for clipping operations
        """
        pc_obj = pc.Pyclipper()
        for path_idx, path in enumerate(self.polylines):
            if np.isin(path_idx, indices):
                path_scaled = pc.scale_to_clipper(path)
                path_scaled = utils.enforce_orientation(path_scaled, path_idx, self.holes)
                pc_obj.AddPath(path_scaled, pc.PT_CLIP)
        return pc_obj

    def make_collection_clipper_offset(self, indices):
        """
        Group given paths/polylines as a clipper offset object for offsetting operations
        """
        pco_obj = pc.PyclipperOffset()
        for path_idx, path in enumerate(self.polylines):
            if np.isin(path_idx, indices):
                path_scaled = pc.scale_to_clipper(path)
                path_scaled = utils.enforce_orientation(path_scaled, path_idx, self.holes)
                pco_obj.AddPath(path_scaled, pc.JT_MITER, pc.ET_CLOSEDPOLYGON)
        return pco_obj

    def add_hole_indices(self, holes):
        """
        Store the indices of paths/polygons that point to holes
        """
        self.holes = np.append(self.holes, holes)

    def add_polylines(self, polylines, line_type='exterior'):
        """
        Add paths/polygons to the object; indicate the type (ext,int,shell,hole etc)
        Returns the indices in which these are stored
        """
        stored_indices = np.empty(0, dtype=int)
        old_size = len(self.polylines)
        new_size = old_size + len(polylines)
        self.polylines.resize(new_size, refcheck=False)
        for i, polyline in enumerate(polylines):
            current_index = old_size + i
            self.polylines[current_index] = np.round(polyline.astype(np.float), prec.p_f_p)
            stored_indices = np.append(stored_indices, current_index)
            self.store_type_index(line_type, current_index)
        return stored_indices

    def gen_3d_polylines(self, fast=True, z=None):
        """
        Interpolate to get the z data of the polylines
        If fast is enabled, interpolation will be done using a precomputed delaunay triangulation for the layer
        """
        print('Generating 3D PolyLines...')
        self.polylines_3d.resize(len(self.polylines), refcheck=False)
        if z is not None:
            for i, polyline in enumerate(self.polylines):
                self.polylines_3d[i] = np.hstack((polyline, np.ones((polyline.shape[0],1))*z))
            return

        if fast:
            self.gen_layer_delaunay()
            for i, polyline in enumerate(self.polylines):
                self.polylines_3d[i] = np.around(utils.interpolate_fast(polyline, self.layer_de, self.xyz_data[:, 2]),
                                                 prec.p_f_p)
        else:
            for i, polyline in enumerate(self.polylines):
                self.polylines_3d[i] = np.around(utils.interpolate_xy(polyline, self.xyz_data), prec.p_f_p)
        self.handle_3d_polylines_nan()

    def gen_layer_delaunay(self):
        """
        Uses scipy to create a delaunay triangulation from the stored xyz data of the layer
        """
        self.layer_de = Delaunay(np.column_stack((self.xyz_data[:, 0], self.xyz_data[:, 1])), qhull_options='QJ Qs')

    def handle_3d_polylines_nan(self):
        """
        Iterates through the stored 3D polylines and reverts to nearest neighbour interpolation if any z data is nan
        """
        for i, polyline_3d in enumerate(self.polylines_3d):
            if np.isnan(self.polylines_3d[i]).any():
                print('\033[31m' + 'Fast interpolation returned nan, reverting to nearest neighbour.' + '\033[0m')
                if self.layer_de is None:
                    self.gen_layer_delaunay()
                self.polylines_3d[i] = np.around(utils.interpolate_nearest(self.polylines[i], self.layer_de,
                                                                        self.xyz_data[:, 2]), prec.p_f_p)

    def store_type_index(self, line_type, index):
        """
        Stores index of path in correct container, given polyline type
        """
        if line_type == 'exterior':
            self.ext_shells = np.append(self.ext_shells, index)
        elif line_type == 'intermediate':
            self.inter_shells = np.append(self.inter_shells, index)
        elif line_type == 'interior':
            self.int_shells = np.append(self.int_shells, index)
        elif line_type == 'infill shell':
            self.infill_shells = np.append(self.infill_shells, index)
        elif line_type == 'infill':
            self.infill = np.append(self.infill, index)
        elif line_type == 'adhesion':
            self.adhesion = np.append(self.adhesion, index)
        else:
            print('Incorrect type given. No index stored')

    def add_to_region_group(self, indices, region_idx):
        """
        Groups paths that are common to the same surface/body
        """
        if region_idx >= self.region_groups.size:
            old_size = len(self.region_groups)
            size_diff = region_idx - old_size
            new_size = old_size+size_diff if size_diff > 0 else old_size+1
            self.region_groups.resize(new_size, refcheck=False)
        if type(self.region_groups[region_idx]) is int:  # for some reason the empty array is initialised as an int
            self.region_groups[region_idx] = indices
        else:
            self.region_groups[region_idx] = np.append(self.region_groups[region_idx], indices)

    def get_indices_in_region(self, region_idx, indices):
        """
        Given an array of indices, return those that belong to the given region
        """
        in_region = np.empty(0, dtype=int)
        for i, index in enumerate(indices):
            if np.isin(index, self.region_groups[region_idx]):
                in_region = np.append(in_region, index)
        return in_region

    def get_int_shells_in_region(self, region_idx):
        """
        Get the indices of polylines that define interior shells for a given region
        """
        return self.get_indices_in_region(region_idx, self.int_shells)

    def get_infill_shells_in_region(self, region_idx):
        """
        Get the indices of polylines that define infill shells for a given region
        """
        return self.get_indices_in_region(region_idx, self.infill_shells)

    def get_ext_shells_in_region(self, region_idx):
        """
        Get the indices of polylines that define exterior shells for a given region
        """
        return self.get_indices_in_region(region_idx, self.ext_shells)