import time
import slicer.utils as utils
import numpy as np
import pyvista as pv
import pyclipper as pc
import trimesh as tm
import pymeshfix as pf
from .PolyLines import PolyLines
from . import Infill
from .Contour import prec


class Slicer:

    def __init__(self, model, settings, repair_first=True):

        if isinstance(model, str):
            try:
                self.model_mesh = pv.read(model)
            except FileNotFoundError:
                raise Exception("Please provide a valid path to a valid stl/obj etc.")
        elif isinstance(model, pv.core.pointset.PolyData):
            self.model_mesh = model

        if repair_first:  # repair in trimesh
            temp_mesh = pf.MeshFix(self.model_mesh)
            temp_mesh = tm.base.Trimesh(vertices=temp_mesh.v, faces=temp_mesh.f)
            tm.repair.fix_inversion(temp_mesh)
            tm.repair.fix_winding(temp_mesh)
            self.model_mesh = pv.wrap(temp_mesh)  # convert trimesh to pyvista

        self.settings = settings
        utils.center_mesh(self.model_mesh, self.settings.bed_centre)
        self.heights = utils.get_model_height_arr(self.model_mesh, self.settings.layer_height)
        self.shell_spacing, self.extrusion_width = self.get_shell_spacing_and_width()
        settings.set_extrusion_width(self.extrusion_width)
        self.top_bottom_pattern = 'hatch'
        self.resample_res = settings.resample_res
        self.t_b_fill_angle = 45
        self.invert = True

    def slice_surf(self, surf):
        """
        Slice a surface
        """
        sliced_surf = surf.clip_surface(self.model_mesh, invert=self.invert)
        return sliced_surf.clean(inplace=False)

    def slice_model(self, surfs):
        """
        Generate sliced layers
        """
        self.surfs = surfs
        s = time.time()
        print('\033[93m Generating slices...\033[0m')
        self.slices = []
        for layer_i, surf in enumerate(self.surfs):
            try:
                self.slices.append(self.slice_surf(surf))
            except:
                # Empty mesh returned - remove remaining surfs and break - slicing is done
                self.surfs = surfs[:layer_i]
                print("Layer {} is empty. Finishing.".format(layer_i, len(surfs)))
                break
            print("Sliced layer {} of {}".format(layer_i, len(surfs)))
        e = time.time()
        print('\033[96m Slicing complete in {} seconds!\033[0m'.format(round(e-s, 3)))

    def get_shell_spacing_and_width(self):
        """
        Return shell spacing. If auto_spacing is true, extrusion width is calculated
        """
        extrusion_width = self.settings.nozzle_diam
        rel_layer_height = self.settings.layer_height
        overlap_factor = 1
        min_width = self.settings.nozzle_diam * 1.05
        max_width = self.settings.nozzle_diam * 1.4
        if self.settings.auto_spacing:
            extrusion_width = ((self.settings.nozzle_diam * self.settings.nozzle_diam) * np.pi
                               + (rel_layer_height * rel_layer_height) * (4 - np.pi)) / (4 * rel_layer_height)
        extrusion_width = min(extrusion_width, max_width)
        extrusion_width = max(extrusion_width, min_width)
        min_shell_spacing = extrusion_width - rel_layer_height * (1 - (np.pi / 4))
        shell_spacing = extrusion_width - overlap_factor*(extrusion_width - min_shell_spacing)
        return np.round(shell_spacing, prec.e_f_p), np.round(extrusion_width, prec.e_f_p)

    def gen_shells(self):
        """
        Generate shells, specify shell spacing and number
        """
        print('\033[93m Generating shells...\033[0m')
        s = time.time()

        self.pl_arr = []
        for layer_idx in range(len(self.slices)):
            print('Generating shells for layer {} of {}'.format(layer_idx, len(self.slices)))
            regions = utils.slice_boundries(self.slices[layer_idx])
            self.pl_arr.append(PolyLines(self.surfs[layer_idx], layer_idx, self.surfs[layer_idx]))
            shells = self.gen_region_shells(regions)
            [self.store_shells(offset_paths, self.pl_arr[layer_idx], region_idx)
                for region_idx, offset_paths in enumerate(shells)]

        e = time.time()
        print('\033[96m Shell generation complete in {} seconds!\033[0m'.format(round(e-s, 3)))

    def gen_region_shells(self, regions):
        """
        Iterates through the regions on a layer, generates and stores offset shells
        """
        shells = []
        for paths in regions:
            pco_obj = pc.PyclipperOffset()
            holes = utils.find_holes(paths)
            for path_idx, path in enumerate(paths):
                path_scaled = pc.scale_to_clipper(path)
                path_scaled = utils.enforce_orientation(path_scaled, path_idx, holes)
                pco_obj.AddPath(path_scaled, pc.JT_MITER, pc.ET_CLOSEDPOLYGON)
            offset_paths = utils.offset_polygons_multi(pco_obj, self.shell_spacing, self.settings.num_shells, self.resample_res,
                                                    direction='negative')
            offset_paths = utils.close_path_nd_arr_implicit(offset_paths, resample_res=self.resample_res)
            shells.append(offset_paths)
        return shells

    def store_shells(self, paths, pl, region_idx):
        """
        Stores shells for a single region into a PolyLines object
        """
        for offset_idx, lines in enumerate(paths):
            if offset_idx == 0:
                line_type = 'exterior'
            elif offset_idx == len(paths) - 1:
                line_type = 'interior'
            else:
                line_type = 'intermediate'
            indices = pl.add_polylines(lines, line_type=line_type)
            pl.add_to_region_group(indices, region_idx)
            pl.add_hole_indices(indices[utils.find_holes(lines)])

    def gen_brim(self):
        """
        Generates a brim by outwardly offsetting the external shells on the bottom layer that are not holes
        """
        num_brim_lines = int(self.settings.brim_width/self.shell_spacing)
        external_shells_indices = self.pl_arr[0].ext_shells
        external_shells = self.pl_arr[0].polylines[external_shells_indices]
        hole_indices = utils.find_holes(external_shells)
        outermost_shells_indices = np.delete(external_shells_indices, hole_indices)
        outermost_shells = self.pl_arr[0].polylines[outermost_shells_indices]  # do not generate brim on holes
        pco_obj = pc.PyclipperOffset()  # generate brim paths by offsetting the external shells
        path_scaled = pc.scale_to_clipper(outermost_shells)
        pco_obj.AddPaths(path_scaled, pc.JT_MITER, pc.ET_CLOSEDPOLYGON)
        brim_paths = utils.offset_polygons_multi(pco_obj, self.shell_spacing, num_brim_lines, resample_res=None,
                                              direction='positive')
        brim_paths = np.delete(brim_paths, 0, 0)  # do not store first path as it is the same as the external shell
        for i, path in enumerate(brim_paths):  # clip brim paths to bounds of external shells
            path_scaled = pc.scale_to_clipper(path)
            first_layer_aabb_obj = pc.Pyclipper()
            first_layer_aabb_obj.AddPaths(pc.scale_to_clipper(outermost_shells), pc.PT_CLIP)
            first_layer_aabb_obj.AddPaths(path_scaled, pc.PT_SUBJECT)
            clipped_brim_paths = utils.union_polygon(first_layer_aabb_obj, resample_res=None)
            first_layer_aabb_obj.Clear()
            self.pl_arr[0].add_polylines(clipped_brim_paths, line_type='adhesion')  # store points not in bb

    def gen_3d_data_for_array(self, fast=True):
        """
        Iterates through the stored PolyLine array and calls the gen_3d_polylines method
        """
        s = time.time()
        print('\033[93m Getting 3D data...\033[0m')
        [layer.gen_3d_polylines(fast) for layer in self.pl_arr]
        e = time.time()
        print('\033[96m 3D data generation complete in {} seconds!\033[0m'.format(round(e-s, 3)))

    def gen_region_infill_bounds(self, pl, region_idx):
        pco_obj = pl.make_collection_clipper_offset(pl.get_int_shells_in_region(region_idx))
        offset_paths = utils.offset_polygons_multi(pco_obj, self.shell_spacing/2, 2, resample_res=None,
                                                direction='negative')
        line_type = 'infill shell'
        for offset_idx, path in enumerate(offset_paths):
            if offset_idx == 0:
                continue  # discard first instance as it is same as internal shell
            indices = pl.add_polylines(path, line_type=line_type)
            pl.add_to_region_group(indices, region_idx)
            pl.add_hole_indices(indices[utils.find_holes(path)])

    def gen_infill(self):
        """
        Generate infill based on specified type
        """
        print('\033[93m Generating infill...\033[0m')
        s = time.time()

        for pl in self.pl_arr:
            if pl.layer_i % 2:
                self.t_b_fill_angle = 135
            else:
                self.t_b_fill_angle = 45
            for region_idx, _ in enumerate(pl.region_groups):
                self.gen_region_infill(pl, region_idx)

        e = time.time()
        print('\033[96m Infill generation complete in {} seconds!\033[0m'.format(round(e-s, 3)))

    def gen_region_infill(self, pl, region_idx):
        """
        Generate infill for a single region
        """
        self.gen_region_infill_bounds(pl, region_idx)
        try:
            ex_bb_obj = pl.make_collection_clipper(pl.get_infill_shells_in_region(region_idx))
        except pc.ClipperException:  # region not valid for infill
            return
        top_bottom = self.check_if_top_bottom(pl, region_idx)

        if top_bottom:
            # too slow to make infill for whole bounding box for this
            ex_bb = utils.paths_bb(ex_bb_obj)
        else:
            # needed for infill allignment acorss layers but slower since whole bounding box is used
            bounds = self.model_mesh.bounds
            ex_bb = np.array([[bounds[0], bounds[2]], [bounds[1], bounds[3]]])

        if top_bottom:
            print("Region {} on layer {} was identified as an external layer.".format(region_idx, pl.layer_i))
            infill = self.gen_top_bottom(ex_bb)
        elif self.settings.infill_pattern == 'honeycomb':
            infill = self.gen_honeycomb_infill(ex_bb)
        elif self.settings.infill_pattern == 'rectilinear':
            if pl.layer_i % 2:
                angle = 0
            else:
                angle = 90
            infill = self.gen_rect_infill(ex_bb, angle=angle)
        else:
            infill = None
        if infill is not None:
            infill_scaled = pc.scale_to_clipper(infill)
            pco_obj = pc.PyclipperOffset()  # add a negligible safety offset to clean up paths before clip
            pco_obj.AddPaths(infill_scaled, pc.JT_MITER, pc.ET_OPENBUTT)
            infill_scaled = pco_obj.Execute(pc.scale_to_clipper(1e-5))
            pco_obj.Clear()
            try:
                ex_bb_obj.AddPaths(infill_scaled, pc.PT_SUBJECT)
                clipped_infill = utils.intersect_polygon(ex_bb_obj, self.resample_res)  # clip infill
                if len(clipped_infill) == 0:  # if no infill is returned, continue
                    return
                unique_infill = utils.remove_duplicate_paths(clipped_infill, prec.p_f_p)
                if top_bottom:
                    sorted_mod2_infill = utils.top_bottom_clean(unique_infill, self.resample_res, self.extrusion_width)
                else:
                    if self.settings.infill_pattern == 'honeycomb':
                        unique_infill = utils.honeycomb_clean(unique_infill)
                    sorted_infill = utils.sort_paths(unique_infill)  # reorder paths
                    sorted_mod2_infill = utils.reverse_mod2(sorted_infill)  # reverse every other path
                indices = pl.add_polylines(sorted_mod2_infill, 'infill')
                pl.add_to_region_group(indices, region_idx)
            except pc.ClipperException:
                return
            ex_bb_obj.Clear()
        return

    def gen_honeycomb_infill(self, ex_bb):
        """
        Generate honeycomb infill for a given bounding box
        """
        hc = Infill.Honeycomb(self.settings.infill_density, ex_bb, self.settings.infill_threshold, self.shell_spacing, self.resample_res)
        return hc.fill_surface()

    def gen_rect_infill(self, ex_bb, angle):
        """
        Generate rectilinear infill
        """
        rc = Infill.Rectilinear(self.settings.infill_density, angle, ex_bb, self.settings.infill_threshold,
                                self.shell_spacing, self.resample_res)
        return rc.fill_surface()

    def gen_top_bottom(self, ex_bb):
        """
        Generate top/bottom layers
        """
        rc = Infill.Rectilinear(1, self.t_b_fill_angle, ex_bb, self.settings.infill_threshold,
                                self.shell_spacing, self.resample_res)
        return rc.fill_surface()

    def check_if_top_bottom(self, pl, region_idx):
        """
        Check if the current region is a top or bottom layer
        """
        if self.settings.top_bottom_shells == 0:
            return False
        top_layer_to_check = self.get_top_layer_check_idx(pl.layer_i)
        bottom_layer_to_check = self.get_bottom_layer_check_idx(pl.layer_i)
        if top_layer_to_check is None or bottom_layer_to_check is None:
            return True

        is_top_bottom = []
        for to_check in [top_layer_to_check, bottom_layer_to_check]:
            other_pl_arr = self.pl_arr[to_check]
            pc_obj = pc.Pyclipper()
            current_paths = pc.scale_to_clipper(pl.polylines[pl.get_ext_shells_in_region(region_idx)])
            other_paths = pc.scale_to_clipper(other_pl_arr.polylines[other_pl_arr.ext_shells])
            #######
            # add a safety offset to other paths so polygon overlap test doesn't return noise
            # this ensures the clip polygon completely envelops the subject polygon
            #######
            try:
                safety_offset = self.extrusion_width*4
                pco_obj = pc.PyclipperOffset()
                pco_obj.AddPaths(other_paths, pc.JT_MITER, pc.ET_CLOSEDPOLYGON)
                other_paths_offset = utils.offset_polygon(pco_obj, safety_offset, resample_res=None)
                other_paths_offset = pc.scale_to_clipper(other_paths_offset)
                pc_obj.AddPaths(current_paths, pc.PT_SUBJECT, False)  # allow open paths
                pc_obj.AddPaths(other_paths_offset, pc.PT_CLIP, True)  # force closed paths
                is_external = not utils.polygon_overlap(pc_obj)
                is_top_bottom.append(is_external)
            except pc.ClipperException:
                print("\033[31m Caught a clipper exception when trying to detect top/bottom layers."
                        "This is typically caused by current layer height and non-planar surface "
                        "settings creating geometry on external surfaces that is invalid for clipping. "
                        "Consider changing layer height or non-planar surface settings. \033[0m")
                print('\033[93m' + "Defaulting to fill region with solid infill." + '\033[0m')
                is_top_bottom.append(True)
        return True if any(is_top_bottom) else False

    def get_top_layer_check_idx(self, layer_idx):
        """
        Returns the index of the layer above the one currently being checked
        This index is used to check for overlap with the current region's polygons
        """
        max_layer_idx = len(self.surfs) - 1
        layer_to_check = layer_idx + self.settings.top_bottom_shells
        if layer_idx >= max_layer_idx or layer_to_check >= max_layer_idx:
            return None
        else:
            return layer_to_check

    def get_bottom_layer_check_idx(self, layer_idx):
        """
        Returns the index of the layer below the one currently being checked
        This index is used to check for overlap with the current region's polygons
        """
        min_layer_idx = 0
        layer_to_check = layer_idx - self.settings.top_bottom_shells
        if layer_idx <= min_layer_idx or layer_to_check <= min_layer_idx:
            return None
        else:
            return layer_to_check

    def slice(self, surfs):
        """
        Call all required methods to slice the model, generating shells and infill
        """
        self.slice_model(surfs)
        self.gen_shells()
        if self.settings.brim_width >= self.shell_spacing:
            self.gen_brim()
        self.gen_infill()
        self.gen_3d_data_for_array()

    def slice2polylines(self, sl, layer_i, lim1_region=False, planar=False):
        """
        convert an pyvista slice to polylines object
        """
        pl = PolyLines(sl, layer_i, self.surfs[layer_i])

        # gen shells
        regions = utils.slice_boundries(sl)
        if len(regions) > 1 and lim1_region:
            print("Warning: Multiple regions detected in slice, only doing 1")
            regions = [regions[0]]
        shells = self.gen_region_shells(regions)
        [self.store_shells(sh, pl, i) for i,sh in enumerate(shells)]

        # gen infill
        # NOTE: must run self.gen_shells() before this but it is fast
        if pl.layer_i % 2:
            self.t_b_fill_angle = 135
        else:
            self.t_b_fill_angle = 45
        [self.gen_region_infill(pl, i) for i in range(len(regions))]

        # get 3d data for polylines
        if planar:
            # assume all points are at same z
            z = self.surfs[layer_i].points[0,2]
            pl.gen_3d_polylines(z=z)
        else:
            pl.gen_3d_polylines()
        return pl