import numpy as np
import pickle
import pymeshfix as pf
import pyvista as pv
import trimesh as tm
import scipy as sc
from scipy.spatial import KDTree
from scipy.interpolate import LinearNDInterpolator,NearestNDInterpolator
import trimesh.base
import shapely as sp
import os
from pathlib import Path


def plot_meshes(meshes, show=True):
    """
    Visualise model meshes, return handle to plotter
    """
    if show:
        off_screen = False
    else:
        off_screen = True
    plotter = pv.Plotter(off_screen=off_screen)
    for i, mesh in enumerate(meshes):
        plotter.add_mesh(mesh, color=np.random.randint(0, 255, size=3).tolist(), show_edges=True, lighting=False,
                         opacity=1, line_width=0.1, edge_color=np.random.randint(0, 255, size=3).tolist())
    if show:
        plotter.show()
    return plotter


def copy_over_array(arrs, dim):
    """
    Utility to duplicate data over array
    """
    new_arrs = np.empty([len(arrs), dim], dtype=object)
    for i, arr in enumerate(arrs):
        for j in range(dim):
            new_arrs[i][j] = arr
    return new_arrs


def to_trimesh(mesh, subdivide=False):
    """
    Converts a mesh into a trimesh format
    Optional: subdivide by 4
    """
    mesh_fix = pf.MeshFix(mesh)
    trimesh_mesh = tm.base.Trimesh(vertices=mesh_fix.v, faces=mesh_fix.f)
    if subdivide:
        trimesh_mesh = trimesh_mesh.subdivide()
    return trimesh_mesh


def path_arr_to_polys(paths):
    """
    Given an array of paths, turn them into an array of polygons
    """
    polygons = np.empty(len(paths), dtype=object)
    for i, path in enumerate(paths):
        polygons[i] = sp.geometry.Polygon(path)
    return polygons


def gen_single_surf(x, y, z):
    """
    Generates mesh of a single surface, given xyz data
    """
    surf_points = np.c_[x.reshape(-1), y.reshape(-1), z.reshape(-1)]
    point_cloud = pv.PolyData(surf_points)
    surf = point_cloud.delaunay_2d()
    return surf


def interpolate_xy(target_xy, interp_data):
    """
    Utility function to interpolate z data for given xy coords
    xyz data of current slice are used for interpolation
    """
    interp_xy = np.column_stack((interp_data[:, 0], interp_data[:, 1]))
    interp_z = interp_data[:, 2]
    target_z = sc.interpolate.griddata(interp_xy, interp_z, target_xy)
    target_z = target_z.transpose()
    target_xyz = np.column_stack((target_xy, target_z))
    return target_xyz


def interpolate_fast(target_xy, interp_data_xy_delaunay, interp_data_z):
    """
    Utility function to do fast interpolation using a precomputed delaunay triangulation
    """
    # interpolator = sc.interpolate.CloughTocher2DInterpolator(interp_data_xy_delaunay, interp_data_z)
    interpolator = LinearNDInterpolator(interp_data_xy_delaunay, interp_data_z)
    target_z = interpolator(target_xy)
    target_z = target_z.transpose()
    target_xyz = np.column_stack((target_xy, target_z))
    return target_xyz


def interpolate_nearest(target_xy, interp_data_xy_delaunay, interp_data_z):
    """
    Nearest neighbour interpolation
    """
    interpolator = NearestNDInterpolator(interp_data_xy_delaunay, interp_data_z)
    target_z = interpolator(target_xy)
    target_z = target_z.transpose()
    target_xyz = np.column_stack((target_xy, target_z))
    return target_xyz


def calc_line_dist(p0, p1):
    """
    Calculates the euclidean distance between two points in any dimension
    """
    return np.sqrt(np.sum((p0-p1)**2))


def rotate_2d_point(point, origin, angle):
    """
    Rotates a 2d point by a given angle about origin
    """
    angle = np.deg2rad(angle)
    return [np.cos(angle) * (point[0] - origin[0]) - np.sin(angle) * (point[1] - origin[1]) + origin[0],
            np.sin(angle) * (point[0] - origin[0]) + np.cos(angle) * (point[1] - origin[1]) + origin[1]]


def rotate_2d_line(line, origin, angle):
    """
    Rotates a 2d line segment by a given angle about origin
    """
    return[rotate_2d_point(line[0], origin, angle), rotate_2d_point(line[1], origin, angle)]


def resample_path(path, resample_res):
    """
    Resamples a line with the given resolution
    """
    return tm.path.traversal.resample_path(path, step=resample_res)


def remove_duplicate_paths(paths, precision):
    """
    Given a list of paths, remove any duplicates (to precision)
    """
    unique_paths = np.empty_like(paths)
    for idx, path in enumerate(paths):  # loop through to remove duplicate paths
        unique, inverse = tm.grouping.unique_rows(path, precision)
        unique_path = np.around(path[unique], precision)
        unique_paths[idx] = unique_path
    return unique_paths


def reverse_mod2(paths):
    """
    Reverses every other path
    """
    paths_out = np.empty_like(paths)
    for idx, path in enumerate(paths):  # loop through to reverse paths
        path_out = path
        if idx % 2:  # reverse every second path for efficient traversal
            path_out = np.flip(path_out, axis=0)
        paths_out[idx] = path_out
    return paths_out


def get_closest_unused_point(p, used, tree, max_dist=10):
    dists, inds = tree.query(p, k=10)
    i=0
    while i<len(inds) and inds[i] in used:
        i+=1

    if i>=len(inds) or dists[i] > max_dist:
        return None, None
    return inds[i], dists[i]


def top_bottom_clean(paths, resample_res, extrusion_width):
    """
    straight lines sometimes are bad - cleanup
    and join paths that are close together
    """
    # remove unnecessary points due to bad clipping
    cleaned_paths_endpoints = []
    for i, path in enumerate(paths):
        min_pos = path[np.argmin(path[:,0])]
        max_pos = path[np.argmax(path[:,0])]
        cleaned_paths_endpoints.append(np.array([min_pos,max_pos]))

    # add points back in using linspace
    cleaned_paths = []
    for i,path in enumerate(cleaned_paths_endpoints):
        if np.linalg.norm(path[0]-path[1]) <1:
            # skip path too short
            continue
        cleaned_paths.append(np.linspace(path[0], path[1], 2+int(np.linalg.norm(path[0]-path[1])/resample_res)))

    p_starts = np.array([c[0] for c in cleaned_paths])
    p_ends = np.array([c[-1] for c in cleaned_paths])

    if len(cleaned_paths) == 0:
        return cleaned_paths

    # join together close paths
    out_paths= []
    start_points_tree = KDTree(p_starts)
    end_points_tree = KDTree(p_ends)
    max_join_dist = extrusion_width*2
    i = 0
    current_path = cleaned_paths[0]
    used = [0]
    while i < len(cleaned_paths):
        # get closest start and end points to end of current_path
        p_start, p_start_dist = get_closest_unused_point(current_path[-1], used, start_points_tree, max_join_dist)
        p_end, p_end_dist= get_closest_unused_point(current_path[-1], used, end_points_tree, max_join_dist)

        # if not paths close enough, add and start new path
        if p_start is None and p_end is None:
            # no close points, add current path to out_poses and start new path
            out_paths.append(current_path)
            while i in used:
                i+=1
            if i >= len(cleaned_paths):
                break
            current_path = cleaned_paths[i]
            used.append(i)
            continue
        # if one if close enough and not in used, figure out which one to add (and flip if needed)
        elif p_start is not None and p_end is None:
            p_to_add = p_start
            flip = False
        elif p_start is None and p_end is not None:
            p_to_add = p_end
            flip = True
        else:
            p_to_add = p_start if p_start_dist < p_end_dist else p_end
            flip = p_start_dist > p_end_dist

        if flip:
            current_path = np.concatenate((current_path, np.flip(cleaned_paths[p_to_add],axis=0)), axis=0)
        else:
            current_path = np.concatenate((current_path, cleaned_paths[p_to_add]), axis=0)
        used.append(p_to_add)
    return np.array(out_paths, dtype=object)


def honeycomb_clean(paths):
    """
    Pyclipper messes up point order sometimes
    this fixes the order
    by sorting the paths in the y-direction
    """
    out_paths = []
    for path in paths:
        # sort by y column
        out_paths.append(path[path[:,1].argsort()])
    return np.array(out_paths, dtype=object)

def sort_paths(paths):
    """
    Given an array of paths, order them according to closest start points
    """
    first_points = np.empty([len(paths), 2])
    for idx, path in enumerate(paths):  # store first point of each path
        first_points[idx] = path[0]
    ordered_indices = np.zeros(len(first_points), dtype=int)
    indices_mask = np.ones(len(first_points), dtype=bool)
    first_points_tree = sc.spatial.KDTree(first_points)
    indices_mask[0] = False  # starting at path 0, omit this as a potential closest path
    for idx, point in enumerate(first_points):  # loop through to order paths
        if idx == len(first_points) - 1:
            continue
        masked_points = first_points[indices_mask]  # construct possible closest point candidates
        masked_tree = sc.spatial.KDTree(masked_points)
        # query candidate points/path indices against the current(sorted) path index
        _, masked_idx = masked_tree.query(first_points[ordered_indices[idx]])
        closest_point = masked_points[masked_idx]  # closest point to first point of current path
        _, closest_idx = first_points_tree.query(closest_point)  # find original index of this point
        ordered_indices[idx + 1] = closest_idx  # add this path index as the next in the list
        indices_mask[closest_idx] = False  # remove this path index from future candidates
    return paths[ordered_indices]  # reorder paths


def closest_dist(mesh_prox, points):
    """
    Queries a mesh, returns the distances to the closest point on the mesh from each of the given points
    """
    dists = mesh_prox.signed_distance(points)
    dists = np.abs(dists)
    # tree = KDTree(mesh.points)
    # dists, idx = tree.query(points)
    return dists


def create_prox_obj(mesh):
    """
    Creates a trimesh proximity query object
    """
    if isinstance(mesh, trimesh.base.Trimesh):
        return tm.proximity.ProximityQuery(mesh)
    else:
        return tm.proximity.ProximityQuery(to_trimesh(mesh, subdivide=False))


def line_mesh_collision(mesh, line_origin, line_end):
    """
    Uses ray tracing to find intersections between a given mesh(es) and a line
    Returns true if a collision occurs
    """
    if type(mesh) is np.ndarray:
        for i, mesh_single in enumerate(mesh):
            points, idc = mesh_single.ray_trace(line_origin, line_end)
            if points.size > 0:
                return True
    else:
        points, idc = mesh.ray_trace(line_origin, line_end)
        if points.size > 0:
            return True
    return False


def create_gcode_file_path(stl_file, output_dir=None):
    """
    Creates the directory/path for the output gcode file
    If output_dir is None, the default directory of gcode_out/ is used
    """
    if output_dir is None:
        gcode_file = os.path.basename(stl_file)
        output_dir = "gcode_out/"+os.path.splitext(gcode_file)[0]+'/'
        Path(output_dir).mkdir(parents=True, exist_ok=True)
        gcode_file = os.path.splitext(gcode_file)[0] + ".gcode"
        gcode_file = output_dir + gcode_file
    else:
        gcode_file = os.path.basename(stl_file)
        Path(output_dir).mkdir(parents=True, exist_ok=True)
        gcode_file = os.path.splitext(gcode_file)[0] + ".gcode"
        gcode_file = output_dir + gcode_file
    return gcode_file, output_dir


def transform_to_point(model_mesh, point):
    """
    Transform mesh to centre it at given point
    """
    delta = np.array(point) - np.array(model_mesh.center)
    transform = tm.transformations.translation_matrix(delta)
    model_mesh.transform(transform)


def center_mesh(model_mesh, bed_center):
    """
    Transform mehs to put the bottom of the model on the build plate
    """
    model_height = model_mesh.bounds[5] - model_mesh.bounds[4]
    bed_center[2] = model_height/2
    transform_to_point(model_mesh, bed_center)


def get_model_height_arr(model_mesh, layer_height):
    """
    Get the z-height increments for the bounds of the model
    Dependent on layer height
    """
    model_min_height, model_max_height = model_mesh.bounds[4], model_mesh.bounds[5]
    model_height = model_max_height - model_min_height
    height_step_num = int(model_height / layer_height)
    heights = np.round(np.linspace(model_min_height+layer_height, model_max_height-layer_height, height_step_num), 3)
    return heights


def contour2pv(c):
    """
    Convert a contour to a pyvista PolyData object
    """
    ps = []
    lines = []
    for i, p in enumerate(c.pos):
        ps.append(np.array(p))
        if i > 0:
            lines += [2, i-1, i]

    ps = np.array(ps)
    c_out = pv.PolyData(ps, lines=lines)
    c_out["Normals"] = c.normals
    return c_out

def contours2pv(contours):
    """
    Convert a list of contours to a pyvista PolyData object
    """
    ps = []
    lines = []
    for c in contours:
        for j, p in enumerate(c.pos):
            ps.append(np.array(p))
            if j > 0:
                lines += [2, len(ps)-2, len(ps)-1]

    ps = np.array(ps)
    c_out = pv.PolyData(ps, lines=lines)
    c_out["Normals"] = np.concatenate([c.normals for c in contours])
    return c_out


def save_contours(cs, fname):
    """
    Save a list of contours to a file
    """
    with open(fname + '.contours', 'wb') as f:
        pickle.dump(cs, f)