import trimesh as tm
import numpy as np
from . import general as gn
import warnings
import pyclipper as pc
import time


def get_closed_connected_entities(path):
    """
    Processes a trimesh path and returns a list of closed, connected entities and ordered vertices
    """
    entity_indices = tm.path.traversal.closed_paths(path.entities, path.vertices)
    entity_indices = np.array(entity_indices, ndmin=1).transpose().reshape(-1)
    dangling = np.array(path.dangling, ndmin=1).transpose().reshape(-1)
    dangling_mask = np.isin(entity_indices, dangling)
    entity_indices = entity_indices[~dangling_mask]
    ordered_path = path
    return entity_indices, ordered_path


def slice_boundries(sl):
    """
    pyvista layer slice to region boundries
    """
    # split into regions
    layer_split = gn.to_trimesh(sl, subdivide=False).split(only_watertight=False)
    boundaries= np.empty(len(layer_split), dtype=object)

    # find boundaries for each region
    for r, region in enumerate(layer_split):
        boundaries[r] = gen_discrete_boundaries(region)
    return boundaries


def gen_discrete_boundaries(mesh):
    """
    Given a mesh, find the outlines, order them, and return the discrete paths
    If the boundaries are coplanar (constant z), first projects them to 2d to simplify and merge colinear segments
    then transforms back to 3d. If they are not coplanar, this operation is passed
    """
    boundaries = mesh.outline()
    try:
        boundaries_2d, _ = boundaries.to_planar()
        boundaries_2d = boundaries_2d.simplify()
        boundaries = boundaries_2d.to_3D()
    except ValueError:
        pass
    boundaries.merge_vertices()
    boundaries.remove_duplicate_entities()
    shell_indices, ordered_boundary = get_closed_connected_entities(boundaries)
    shells = np.empty(len(shell_indices), dtype=object)
    for shell_idx, index_val in enumerate(shell_indices):
        shells[shell_idx] = ordered_boundary.discretize_path(np.array(index_val, ndmin=1)).view(np.ndarray)
    return shells


def get_longest_path(paths):
    """
    Given a list of paths, return the one that is the longest
    """
    longest_path_dist = 0
    longest_path_idx = 0
    for i, path in enumerate(paths):
        diff = np.diff(path, axis=0) ** 2
        length = (np.dot(diff, [1] * path.shape[1]) ** 0.5).sum()
        if length > longest_path_dist:
            longest_path_dist = length
            longest_path_idx = i
    return paths[longest_path_idx]


def paths_bb(paths_obj):
    """
    Given a list of paths as a clipper object, return the bounding box encompassing them
    """
    bounding_box = paths_obj.GetBounds()
    x_min = pc.scale_from_clipper(bounding_box.left)
    x_max = pc.scale_from_clipper(bounding_box.right)
    y_min = pc.scale_from_clipper(bounding_box.top)
    y_max = pc.scale_from_clipper(bounding_box.bottom)
    return np.array([[x_min, y_min], [x_max, y_max]])


def close_path_nd_arr_implicit(nd_paths, resample_res=None):
    """
    Iterates through an nd array of paths and closes them if needed
    """
    temp_nd_paths = np.empty(len(nd_paths), dtype=object)
    for p, paths in enumerate(nd_paths):
        temp_nd_paths[p] = close_path_arr_implicit(paths, resample_res)
    return temp_nd_paths


def close_path_arr_implicit(paths, resample_res=None):
    """
    Iterates through an array of paths and closes them if needed
    """
    temp_paths = np.empty(len(paths), dtype=object)
    for p, path in enumerate(paths):
        temp_paths[p] = close_path_implicit(path, resample_res)
    return temp_paths


def close_path_implicit(path, resample_res=None):
    """
    If the tuples at the first and last indices of a path are not equal, the path will be implicitly closed
    by copying the first tuple to the last index and resampling in between
    If the indices already match, returns the original path
    """
    first_point = path[0]
    last_point = path[len(path)-1]
    if not np.allclose(first_point, path[len(path)-1]):
        closing_dist = gn.calc_line_dist(first_point, last_point)
        if resample_res is not None:
            num_steps = int(closing_dist / resample_res)
            closing_path = np.linspace(last_point, first_point, num_steps)
        else:
            closing_path = np.array([first_point])
        path = np.append(path, closing_path, axis=0)
    return path


def find_holes(paths):
    """
    Given an array of paths, finds the indices of those that are holes
    """
    polygons = gn.path_arr_to_polys(paths).tolist()
    roots, edges = tm.path.polygons.enclosure_tree(polygons)  # edges indicate a polygon is contained by another
    holes = np.empty(len(edges.edges), dtype=int)
    edges = [e for e in edges.edges]
    for i, edge in enumerate(edges):
        for j, root in enumerate(roots):  # iterate through all roots
            if np.isin(root, edge):  # find which root the hole is enclosed by
                root_index = edge.index(root)  # get the index of the root from the edge tuple
                holes[i] = edge[~root_index]  # hole is then the node at the other index of the tuple
    return holes


def enforce_orientation(path, path_idx, holes):
    """
    If path index is a hole, ensure orientation is false (cw)
    If path index is an outer path, ensure orientation is true (ccw)
    Returns path with correct orientation
    """
    path_orient = pc.Orientation(path)  # returns true if path is ccw
    if np.isin(path_idx, holes):
        if path_orient:
            path = pc.ReversePath(path)
    else:
        if not path_orient:
            path = pc.ReversePath(path)
    return path


def offset_polygon(pco_obj, offset, resample_res):
    """
    Offsets a 2d polygon
    Accepts the pyclipper object to perform offset on
    Returns array of points defining 2d polygon
    """
    polygon_2d_offset = pco_obj.Execute(pc.scale_to_clipper(offset))
    return process_clipper_results(polygon_2d_offset, resample_res, reject_noise=True)


def process_clipper_results(clipper_output, resample_res, reject_noise=False):
    """
    Comprehends the result paths/polygons from a clipper operation
    Resamples and rejects noise if specified
    """
    clipper_polygons = pc.CleanPolygons(clipper_output)
    clipper_polygons = pc.scale_from_clipper(clipper_polygons)
    polygons = np.array([np.array(paths) for paths in clipper_polygons], dtype=object)
    output_polygons = np.empty(len(polygons), dtype=object)
    noise = None
    if reject_noise and resample_res is not None:
        noise = np.empty(0, dtype=int)
    for path_idx, path in enumerate(polygons):
        if resample_res is None:
            output_polygons[path_idx] = np.array(path, dtype=float)
            continue
        path_resampled = gn.resample_path(path, resample_res)  # resample path for z integrity
        output_polygons[path_idx] = path_resampled
        if len(path_resampled) < 3 and noise is not None:  # reject paths with less than three points (noise)
            noise = np.append(noise, path_idx)
    return output_polygons if noise is None else np.delete(output_polygons, noise, axis=0)


def intersect_polygon(pc_obj, resample_res):
    """
    Accepts the pyclipper object to perform intersection on
    Returns array of points defining 2d polygon
    """
    polygon_2d_clipped = pc_obj.Execute(pc.CT_INTERSECTION, subj_fill_type=pc.PFT_NONZERO,
                                        clip_fill_type=pc.PFT_NONZERO)
    return process_clipper_results(polygon_2d_clipped, resample_res, reject_noise=False)


def union_polygon(pc_obj, resample_res):
    """
    Accepts the pyclipper object to perform union on
    Returns array of points defining 2d polygon
    """
    polygon_2d_clipped = pc_obj.Execute(pc.CT_UNION, subj_fill_type=pc.PFT_NONZERO,
                                        clip_fill_type=pc.PFT_NONZERO)
    return process_clipper_results(polygon_2d_clipped, resample_res, reject_noise=False)


def polygon_overlap(pc_obj):
    """
    Given a pyclipper object, check to see if any overlap exists between the subject and clip polygons
    Returns True if overlap exists
    """
    polygon_clipped = pc_obj.Execute2(pc.CT_DIFFERENCE, subj_fill_type=pc.PFT_EVENODD, clip_fill_type=pc.PFT_EVENODD)
    polygon_clipped = pc.PolyTreeToPaths(polygon_clipped)
    polygon_clipped = pc.scale_from_clipper(polygon_clipped)
    polygon_clipped = np.array([np.array(paths) for paths in polygon_clipped], dtype=object)
    return True if polygon_clipped.size == 0 else False


def offset_polygons_multi(pco_obj, offset_dist, num_offsets, resample_res, direction='negative'):
    """
    Offsets multiple polygons and returns the (merged) result for each offset index
    """
    if direction != 'negative' and direction != 'positive':
        warnings.warn('Invalid offset direction provided. Using default of negative')
    offset = -abs(offset_dist)
    offset = abs(offset) if direction == 'positive' else offset
    offset_polygons = np.empty(num_offsets, dtype=object)
    for offset_idx in range(num_offsets):
        polygon_2d_offset = offset_polygon(pco_obj, offset * offset_idx, resample_res=resample_res)
        offset_polygons[offset_idx] = polygon_2d_offset
    return offset_polygons