import numpy as np
import time
import pyvista as pv
import sympy as sp
from .general import copy_over_array, gen_single_surf


def get_model_xy_grid(model_mesh, resolution, bigger=True):
    """
    Get the xy grid used to generate the surfaces for slicing
    resolution: the number of points on the longer side
    """
    # get bounds and resolution for x and y
    min_x, max_x = model_mesh.bounds[:2]
    min_y, max_y = model_mesh.bounds[2:4]
    x_len = abs(max_x - min_x)
    y_len = abs(max_y - min_y)
    if x_len >= y_len:
        res_x = resolution
        res_y = int(resolution*y_len/x_len)
    else:
        res_y = resolution
        res_x = int(resolution*x_len/y_len)

    # gen grid
    if bigger:
        x = np.linspace(min_x-1e-3, max_x+1e-3, res_x)
        y = np.linspace(min_y-1e-3, max_y+1e-3, res_y)
    else:
        x = np.linspace(min_x, max_x, res_x)
        y = np.linspace(min_y, max_y, res_y)
    x, y = np.meshgrid(x, y)
    return x, y


def planar_z_arr(x, y, heights):
    """
    Generate z data for planar layers
    """
    z_arr = np.empty(len(heights), dtype=object)
    for i in range(len(heights)):
        z = np.full([len(x), len(y)], heights[i])
        z_arr[i] = z
    return z_arr


def non_planar_z_arr(x, y, heights, coeffs=None):
    """
    Generate z data for non-planar layers
    """
    if coeffs is None:
        coeffs = [0.9, 0.2, 0.2]
    z_arr = np.empty(len(heights), dtype=object)
    fade = 0
    fade_factor = 0.04
    for i in range(len(heights)):
        fade_corrected = fade if fade < 1 else 1
        if i == 0 | i == len(heights) - 1:
            z = np.full([len(x), len(y)], heights[i])
        else:
            z = fade_corrected * non_planar_surf_eqn(x, y, coeffs, evaluate=True) + heights[i]
        z_arr[i] = z
        if i <= len(heights) / 2:
            fade += fade_factor
        else:
            fade -= fade_factor
    return z_arr


def non_planar_surf_eqn(nx, ny, coeffs, evaluate=True):
    """
    Equation template for nonplanar surfaces
    If evaluate is True, the function will be evaluated with the given data
    If False, the symbolic equation will be returned
    """
    x, y, a, b, c = sp.symbols('x y a b c', real=True)
    z = a * sp.sin(b * x) * sp.sin(c * y)
    if not evaluate:
        return z
    else:
        z = sp.lambdify([x, y, a, b, c], z)
        return z(nx, ny, coeffs[0], coeffs[1], coeffs[2])


def gen_planar_surf(model_mesh, resolution, z):
    """
    Create a slicing surface
    """
    x, y = get_model_xy_grid(model_mesh, resolution)
    z = np.full(x.shape, z)
    surf_points = np.c_[x.reshape(-1), y.reshape(-1), z.reshape(-1)]

    point_cloud = pv.PolyData(surf_points)
    surf = point_cloud.delaunay_2d()
    return surf


def gen_sine_surfs(model_mesh, resolution, heights, coeffs=None):
    """
    Generate layers
    """
    print('\033[93m' + "Generating surfaces..." + '\033[0m')
    s = time.time()

    x, y = get_model_xy_grid(model_mesh, resolution)
    z = non_planar_z_arr(x, y, heights, coeffs)
    [x, y] = copy_over_array([x, y], len(heights))
    surfs = []
    for l in range(len(heights)):
        surfs.append(gen_single_surf(x[l], y[l], z[l]))
        print("Surface {} added. At Z {} of {}".format(l, heights[l], heights[-1]))

    e = time.time()
    print('\033[96m' + "Surface generation complete in ", round(e-s, 3), " seconds!" + '\033[0m')
    return surfs


def gen_hemisphere_surfs(model_mesh, resolution, roundness, z_max, z_start=1):
    """
    model_mesh: pyvista mesh
    resolution: number of points in longer dimension
    roundness: how curved the top layer is (more is rounder)
    z_max: max height difference allowed between layers
    """
    x, y = get_model_xy_grid(model_mesh, resolution)
    c = [x[0, x.shape[1]//2], y[y.shape[0]//2, 0]] # x-y center
    r = ((x[0,0]-x[-1,-1])**2 + (y[0,0]-y[-1,-1])**2)**0.5
    r = r/roundness
    z_top= ((x-c[0])**2+(y-c[1])**2)**0.5 * roundness
    z_top = z_top.max() - z_top + model_mesh.bounds[5]

    return gen_inbetween_surfs([gen_single_surf(x, y, z_top)], z_start, z_max)


def gen_between_2_surfs(surf1,surf2,z_max):
    """
    Generate surfaces between two waypoint surfaces by linearly interpolating
    """
    z_top = surf2.points[:,2]
    z_start = surf1.points[:,2]
    x = surf1.points[:,0]
    y = surf1.points[:,1]

    n_layers = int((z_top-z_start).max()/z_max) + 1
    surfs = []
    z_mids = np.linspace(0, 1, n_layers)
    for i, z_m in enumerate(z_mids):
        if i==0 or i==n_layers-1:
            continue
        z = z_start + (z_top-z_start)*z_m
        surfs.append(gen_single_surf(x, y, z))
        print("Surface {} of {} added".format(i,len(z_mids)))
    return [surf1] + surfs + [surf2]


def gen_inbetween_surfs(surf_waypoints, z_start, z_max):
    """
    surf_waypoints is a list of surfaces to linearly interpolate between
    """
    surfs = []
    x = surf_waypoints[0].points[:,0]
    y = surf_waypoints[0].points[:,1]
    z_start = np.ones(y.shape)*z_start
    surfs.append(gen_single_surf(x, y, z_start))
    for surf in surf_waypoints:
        surfs += gen_between_2_surfs(surfs[-1],surf,z_max)[1:]
    return surfs


def gen_curved_surfs(model_mesh, resolution, roundness=0.003, z_max=1, z_start=1, axis=0):
    """
    model_mesh: pyvista mesh
    resolution: number of points in longer dimension
    roundness: how curved the top layer is (more is rounder)
    z_max: max height difference allowed between layers
    axis: whether curve is in x-axis or y-axis
    """
    x, y = get_model_xy_grid(model_mesh, resolution)
    c = [x[0, x.shape[1]//2], y[y.shape[0]//2, 0]] # x-y center
    r = ((x[0,0]-x[-1,-1])**2 + (y[0,0]-y[-1,-1])**2)**0.5
    r = r/roundness
    if axis==0:
        z_top= ((x-c[0])**2) * roundness
        z_top = z_top.max() - z_top + model_mesh.bounds[5]
    else:
        z_top= ((y-c[1])**2) * roundness
        z_top = z_top.max() - z_top + model_mesh.bounds[5]


    n_layers = int((z_top.max()-z_start)/z_max) + 1
    surfs = []
    z_mids = np.linspace(z_start, z_top.max(), n_layers)
    z_0 = np.ones(z_top.shape)*z_start
    for i, z_m in enumerate(z_mids):
        z = z_start + (z_top-z_start)*(z_m-z_start)/(z_top.max() - z_start)
        surfs.append(gen_single_surf(x, y, z))
        print("Surface {} of {} added".format(i,len(z_mids)))
    return surfs


def gen_grid_surf(surf_mesh, model_mesh, resolution):
    """
    Generate a structured grid from a surface mesh
    """
    surf_mesh = surf_mesh.extract_surface()
    xall, yall = get_model_xy_grid(model_mesh, resolution, bigger=False)
    zs = []

    for x,y in zip(xall.flatten(), yall.flatten()):
        ip, _ = surf_mesh.ray_trace((x,y,-1000), (x,y,1000), first_point=True)
        if len(ip)==0:
            ip = [x,y,surf_mesh.points.mean(axis=0)[2]]
        zs.append(ip[2])

    zall = np.array(zs).reshape(xall.shape)
    return gen_single_surf(xall,yall,zall)